 /***************************************************************************\
|*                                                                           *|
|*       Copyright 1993-1999 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NOTICE TO USER:   The source code  is copyrighted under  U.S. and     *|
|*     international laws.  Users and possessors of this source code are     *|
|*     hereby granted a nonexclusive,  royalty-free copyright license to     *|
|*     use this code in individual and commercial software.                  *|
|*                                                                           *|
|*     Any use of this source code must include,  in the user documenta-     *|
|*     tion and  internal comments to the code,  notices to the end user     *|
|*     as follows:                                                           *|
|*                                                                           *|
|*       Copyright 1993-1999 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NVIDIA, CORPORATION MAKES NO REPRESENTATION ABOUT THE SUITABILITY     *|
|*     OF  THIS SOURCE  CODE  FOR ANY PURPOSE.  IT IS  PROVIDED  "AS IS"     *|
|*     WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND.  NVIDIA, CORPOR-     *|
|*     ATION DISCLAIMS ALL WARRANTIES  WITH REGARD  TO THIS SOURCE CODE,     *|
|*     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, NONINFRINGE-     *|
|*     MENT,  AND FITNESS  FOR A PARTICULAR PURPOSE.   IN NO EVENT SHALL     *|
|*     NVIDIA, CORPORATION  BE LIABLE FOR ANY SPECIAL,  INDIRECT,  INCI-     *|
|*     DENTAL, OR CONSEQUENTIAL DAMAGES,  OR ANY DAMAGES  WHATSOEVER RE-     *|
|*     SULTING FROM LOSS OF USE,  DATA OR PROFITS,  WHETHER IN AN ACTION     *|
|*     OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,  ARISING OUT OF     *|
|*     OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOURCE CODE.     *|
|*                                                                           *|
|*     U.S. Government  End  Users.   This source code  is a "commercial     *|
|*     item,"  as that  term is  defined at  48 C.F.R. 2.101 (OCT 1995),     *|
|*     consisting  of "commercial  computer  software"  and  "commercial     *|
|*     computer  software  documentation,"  as such  terms  are  used in     *|
|*     48 C.F.R. 12.212 (SEPT 1995)  and is provided to the U.S. Govern-     *|
|*     ment only as  a commercial end item.   Consistent with  48 C.F.R.     *|
|*     12.212 and  48 C.F.R. 227.7202-1 through  227.7202-4 (JUNE 1995),     *|
|*     all U.S. Government End Users  acquire the source code  with only     *|
|*     those rights set forth herein.                                        *|
|*                                                                           *|
 \***************************************************************************/

#ifndef _NV_LINUX_H_
#define _NV_LINUX_H_

#include "nv.h"

#ifdef __KERNEL__

#include <linux/config.h>
#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif

#include <linux/version.h>
#include <linux/wait.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/malloc.h>    /* kmalloc().. */
#include <linux/vmalloc.h>   /* vmalloc().. */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 0)
	#error "This driver does not support 2.1 or earlier kernels"
#else
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 3, 0)
    #define KERNEL_2_2
  #else
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
	  #error "This driver does not support 2.3 kernels"
	#else
	  #define KERNEL_2_4
	#endif // < 2.4
  #endif  // < 2.3
#endif  // < 2.2

#include <linux/autoconf.h>

#ifndef NOAGPGART
#  ifdef CONFIG_AGP
#    define AGPGART
#  endif
#  ifdef CONFIG_AGP_MODULE
#    define AGPGART
#  endif
#endif

#ifdef AGPGART
#include <linux/types.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <linux/agp_backend.h>  // agpgart backend
#include <linux/agpgart.h>      // agpgart
#endif

#if defined(KERNEL_2_4)
    #define GET_MAP_NR(phys_page)      virt_to_page(__va(phys_page))
    #define MEM_MAP_READ_COUNT(map_nr) (atomic_read(&map_nr->count))
    #define MEM_MAP_INC_COUNT(map_nr)  (atomic_inc(&map_nr->count))
    #define MEM_MAP_DEC_COUNT(map_nr)  (atomic_dec(&map_nr->count))
    #define GET_EVENT_QUEUE(nv)        (struct __wait_queue_head *) \
                                           (nv->event_queue)
#else
    #define GET_MODULE_SYMBOL(mod, sym) (void*) get_module_symbol(mod, sym)
    #define GET_MAP_NR(phys_page)      MAP_NR(__va(phys_page))
    #define MEM_MAP_READ_COUNT(map_nr) (atomic_read(&mem_map[map_nr].count))
    #define MEM_MAP_INC_COUNT(map_nr)  (atomic_inc(&mem_map[map_nr].count))
    #define MEM_MAP_DEC_COUNT(map_nr)  (atomic_dec(&mem_map[map_nr].count))
    #define GET_EVENT_QUEUE(nv)        (struct wait_queue **) &(nv->event_queue)
#endif

#ifndef THIS_MODULE
#define THIS_MODULE  (&__this_module)
#endif

#ifndef GET_MODULE_SYMBOL
#include <linux/spinlock.h>  // assume kernel version 2.4
extern spinlock_t unload_lock;
#define NV_GET_MODULE_SYMBOL(sym_value, mod, sym_name) \
    { \
        struct module *mp = THIS_MODULE; \
        struct module_symbol *sym; \
        int i; \
        spin_lock(&unload_lock); \
        if (MOD_CAN_QUERY(mp) && (mp->nsyms > 0)) { \
            for (i = mp->nsyms, sym = mp->syms; i > 0; --i, ++sym) { \
                if (strcmp(sym->name, sym_name) == 0) { \
                    sym_value = sym->value; \
                    break; \
                } \
            } \
        } \
        spin_unlock(&unload_lock); \
    }
#else
#define NV_GET_MODULE_SYMBOL(value, mod, symbol) \
    value = GET_MODULE_SYMBOL(mod, symbol)
#endif

/*
 * An allocated bit of memory using NV_MEMORY_ALLOCATION_OFFSET
 *   looks like this in the driver
 */

typedef struct nv_alloc_s {
    struct nv_alloc_s *next;    
    struct vm_area_struct *vma;
    unsigned int   process_id;
    unsigned int   num_pages;
    void         **page_table;          /* list of physical pages allocated */
    void          *agp_mapping;         /* ptr to linear agp mapping of pages */
    /* for nvagp */
    unsigned int   class;
    void          *priv_data;
} nv_alloc_t;

/* linux-specific version of old nv_state_t */
/* this is a general os-specific state structure. the first element *must* be
   the general state structure, for the generic unix-based code */
typedef struct {
    nv_state_t nv_state;

    nv_alloc_t *alloc_queue;

    // bottom half interrupt handler info; per device
    struct tq_struct *bh;

    U032 vblank_notifier;
    U032 waiting_for_vblank;

    /* queue for for NV's OS events */
    void *event_queue;

} nv_linux_state_t;


/*
 * poll(2) support for event notifiers
 *
 * A notifer can be turned into an "OS" event.
 * On UNIX, this means you can block using poll() while waiting for any
 * of your OS events to fire.
 *
 * When poll() on the NV file descriptor notices that you have one or more
 * events, it returns POLLPRI.  At that time it clears the count of fired
 * notifiers.
 *
 * Upon return from poll() just poll your notifiers.
 */

/*
 * Per-file private data structure.
 * We keep stuff here that is separate on a per-file basis.
 */

typedef struct {
    U032 any_fired_notifiers;
    void *nv_ptr;
} nv_file_private_t;

/* for the card devices */
#define NV_GET_NVL_FROM_FILEP(filep) \
    (nv_linux_state_t*) (((nv_file_private_t *)((filep)->private_data))->nv_ptr)

/* for the control device */
#define NV_GET_NV_FROM_FILEP(filep) \
    (nv_state_t*) (((nv_file_private_t *)((filep)->private_data))->nv_ptr)

#define NV_HIDE_IN_FILEP(filep, nvp) \
    (((nv_file_private_t *)((filep)->private_data))->nv_ptr = (void *) (nvp))
    

#define NV_GET_NVL_FROM_NV_STATE(nv) \
    ((nv_linux_state_t *) nv->os_state)

#define NV_STATE_PTR(nvl)   (&((nvl)->nv_state))

/* all of the below kernel calls could sleep.
 * if we sleep with the nv lock, we could hang an incoming interrupt
 * and lock the system. So, drop the lock, perform the call, and
 * regrab the lock.
 */
#if defined(IA64)
#define NV_VMALLOC(ptr, size) \
    { \
        NVRM_LOCK_SETUP(); \
        NVRM_LOCK_RELEASE(); \
        (void *) ptr = vmalloc_32(size); \
        NVRM_LOCK_GRAB(); \
    }
#else
#define NV_VMALLOC(ptr, size) \
    { \
        NVRM_LOCK_SETUP(); \
        NVRM_LOCK_RELEASE(); \
        (void *) ptr = vmalloc(size); \
        NVRM_LOCK_GRAB(); \
    }
#endif

#define NV_VFREE(ptr) \
    { \
        NVRM_LOCK_SETUP(); \
        NVRM_LOCK_RELEASE(); \
        vfree((void *) ptr); \
        NVRM_LOCK_GRAB(); \
    }

/* only use this because GFP_KERNEL may sleep..
 * GFP_ATOMIC is ok, it won't sleep
 */
#define NV_KMALLOC(ptr, size) \
    { \
        NVRM_LOCK_SETUP(); \
        NVRM_LOCK_RELEASE(); \
        (void *) ptr = kmalloc(size, GFP_KERNEL); \
        NVRM_LOCK_GRAB(); \
    }

#define NV_KFREE(ptr) \
    { \
        NVRM_LOCK_SETUP(); \
        NVRM_LOCK_RELEASE(); \
        kfree((void *) ptr); \
        NVRM_LOCK_GRAB(); \
    }



#endif  /* __KERNEL__ */
#endif  /* _NV_LINUX_H_ */
