 /***************************************************************************\
|*                                                                           *|
|*       Copyright 1993-2000 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NOTICE TO USER:   The source code  is copyrighted under  U.S. and     *|
|*     international laws.  Users and possessors of this source code are     *|
|*     hereby granted a nonexclusive,  royalty-free copyright license to     *|
|*     use this code in individual and commercial software.                  *|
|*                                                                           *|
|*     Any use of this source code must include,  in the user documenta-     *|
|*     tion and  internal comments to the code,  notices to the end user     *|
|*     as follows:                                                           *|
|*                                                                           *|
|*       Copyright 1993-2000 NVIDIA, Corporation.  All rights reserved.      *|
|*                                                                           *|
|*     NVIDIA, CORPORATION MAKES NO REPRESENTATION ABOUT THE SUITABILITY     *|
|*     OF  THIS SOURCE  CODE  FOR ANY PURPOSE.  IT IS  PROVIDED  "AS IS"     *|
|*     WITHOUT EXPRESS OR IMPLIED WARRANTY OF ANY KIND.  NVIDIA, CORPOR-     *|
|*     ATION DISCLAIMS ALL WARRANTIES  WITH REGARD  TO THIS SOURCE CODE,     *|
|*     INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY, NONINFRINGE-     *|
|*     MENT,  AND FITNESS  FOR A PARTICULAR PURPOSE.   IN NO EVENT SHALL     *|
|*     NVIDIA, CORPORATION  BE LIABLE FOR ANY SPECIAL,  INDIRECT,  INCI-     *|
|*     DENTAL, OR CONSEQUENTIAL DAMAGES,  OR ANY DAMAGES  WHATSOEVER RE-     *|
|*     SULTING FROM LOSS OF USE,  DATA OR PROFITS,  WHETHER IN AN ACTION     *|
|*     OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,  ARISING OUT OF     *|
|*     OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOURCE CODE.     *|
|*                                                                           *|
|*     U.S. Government  End  Users.   This source code  is a "commercial     *|
|*     item,"  as that  term is  defined at  48 C.F.R. 2.101 (OCT 1995),     *|
|*     consisting  of "commercial  computer  software"  and  "commercial     *|
|*     computer  software  documentation,"  as such  terms  are  used in     *|
|*     48 C.F.R. 12.212 (SEPT 1995)  and is provided to the U.S. Govern-     *|
|*     ment only as  a commercial end item.   Consistent with  48 C.F.R.     *|
|*     12.212 and  48 C.F.R. 227.7202-1 through  227.7202-4 (JUNE 1995),     *|
|*     all U.S. Government End Users  acquire the source code  with only     *|
|*     those rights set forth herein.                                        *|
|*                                                                           *|
 \***************************************************************************/

#include <linux/config.h>

#if defined(CONFIG_SMP) && !defined(__SMP__)
#define __SMP__
#endif


// pull in the versioned kernel api defines if needed
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>

#include <nv-linux.h>			        // needs to precede other headers (SMP)
#define RM_HEADERS_MINIMAL      // we just need top half of nvrm.h, not the entire
#include <nvrm.h>               // tree of include files

#include <linux/stddef.h>
#include <linux/mm.h>
#include <linux/kernel.h> 		// printk()
#include <linux/fs.h>     		// everything...
#include <linux/errno.h>  		// error codes
#include <linux/types.h>  		// size_t
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        	// O_ACCMODE
#include <linux/pci.h>    		// PCI stuff
#include <linux/delay.h>  		// udelay()
#include <linux/interrupt.h>           
#include <linux/tqueue.h>               // struct tq_struct 
#include <linux/poll.h>
#ifdef CONFIG_PM_NOT_YET
#include <linux/pm.h>                   // power management
#endif

#include <asm/page.h>
#include <asm/pgtable.h>  		// pte bit definitions
#include <asm/system.h>                 // cli(), *_flags
#include <asm/segment.h>                // memcpy and verify_write
#include <asm/uaccess.h>		// access_ok
#include <asm/io.h>
#include <asm/softirq.h>                // mark_bh
#ifdef KERNEL_2_4
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif
#include <asm/smp.h>

#include <linux/wrapper.h>

#include <nv_ref.h>

#include <os-interface.h>              // KernInitAgp

#define LINUX_VMA_DEV(vma)  ((vma)->vm_file->f_dentry->d_inode->i_rdev)

#ifdef KERNEL_2_4
#define LINUX_VMA_OFFS(vma) (((vma)->vm_pgoff) << PAGE_SHIFT)
#else
#define LINUX_VMA_OFFS(vma) ((vma)->vm_offset)
#endif

#define NV_PAGE_ALIGN(addr)    ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)  ( (addr) & (PAGE_SIZE - 1) )

/* 
 * Our main lock for everything.
 */

spinlock_t nv_lock;
unsigned long * volatile nv_locked;
unsigned long volatile nv_locked_by;  /* pid, or '-1' for interrupt */

/*
 * our global state; one per device
 */

nv_linux_state_t nv_linux_devices[NV_MAX_DEVICES];

#ifdef CONFIG_PM_NOT_YET
/* XXX PM do we only need one, or one for each device? */
static struct pm_dev *pm_nv_dev;
#endif

/*
 * And one for the control device
 */

nv_linux_state_t nv_ctl_device;

// keep track of opened clients and their process id so they
//   can be free'd up on abnormal close
nv_client_t       nv_clients[NV_MAX_CLIENTS];
struct tq_struct  nv_bottom_halves[NV_MAX_CLIENTS];

#ifdef CONFIG_PROC_FS
/*
 * Global proc entry for /proc/nv 
 */
 
struct proc_dir_entry *nv_proc_entry;
#endif /* CONFIG_PROC_FS */

/*
 * pick apart our minor device number
 * low 3 bits is NV device
 * if 255, then its the control device
 */

#define NV_DEVICE_NUMBER(_minor) ((_minor) & 0x0f)
#define NV_DEVICE_IS_CONTROL_DEVICE(_minor) \
             (((_minor) & 0xFF) == 0xFF)

/*
 * Hardware access macros for the kernel driver only
 * The resource manager part of the driver uses different values here
 */

// HW register access macros
#define REG_WR08(a,d)   nv->regs_u->Reg008[(a)]   = (U008) (d)
#define REG_WR16(a,d)   nv->regs_u->Reg016[(a)/2] = (U016) (d)
#define REG_WR32(a,d)   nv->regs_u->Reg032[(a)/4] = (U032) (d)

#define REG_RD08(a)     nv->regs_u->Reg008[(a)]
#define REG_RD16(a)     nv->regs_u->Reg016[(a)/2]
#define REG_RD32(a)     nv->regs_u->Reg032[(a)/4]



#ifdef DEBUG
#define NV_DMSG(nv, fmt, args...)  NV_iDMSG(nv, fmt, ## args)
#else
#define NV_DMSG(nv, fmt, args...)
#endif


/* functions in core resman that we need to call */
void NvUpdateAGPConfig(PHWINFO);
BOOL RmInitAGP(PHWINFO);
BOOL RmTeardownAGP(PHWINFO);
RM_STATUS NvAllocAGPPages ( PHWINFO, VOID **, U032, U032, VOID **);
RM_STATUS NvMapAGPPages ( PHWINFO, VOID **, U032, VOID *);
RM_STATUS NvFreeAGPPages ( PHWINFO, VOID **, VOID * );

/***
 *** STATIC functions, only in this file
 ***/

/* nvos_ functions.. do not take a state device parameter  */
static int           nvos_is_nv_device(unsigned char, unsigned char, 
                                       unsigned short, unsigned short);  
static int           nvos_set_primary_card(nv_ioctl_primary_card_t *info);
static void          nvos_probe_devices(void);
static unsigned long nvos_get_kvirt_address(unsigned long address);
static void         *nvos_malloc(unsigned long);
static void          nvos_free(void **);

#ifdef CONFIG_PROC_FS
static void          nvos_proc_create(void);
static void          nvos_proc_remove(void);
#endif /* CONFIG_PROC_FS */

/* nvl_ functions.. take a linux state device pointer */
static nv_alloc_t   *nvl_find_alloc(nv_linux_state_t *, unsigned long, nv_alloc_t **);
static nv_alloc_t   *nvl_find_agp_alloc(nv_linux_state_t *, unsigned long, nv_alloc_t **);
static nv_alloc_t   *nvl_find_alloc_by_num_pages(nv_linux_state_t *, unsigned long, nv_alloc_t **);

/* nv_ functions. not used in main resman, but take general state device */
static void          nv_interrupts_disable(nv_state_t *);
static void          nv_interrupts_enable(nv_state_t *);
static int           nv_map_device(nv_state_t *);
static void          nv_unmap_device(nv_state_t *);

/***
 *** EXPORTS to Linux Kernel
 ***/

/* linux module interface functions (called by linux kernel) */
int           init_module(void);
void          cleanup_module(void);

/* nv_kern_ functions, interfaces used by linux kernel */
void          nv_kern_vma_open(struct vm_area_struct *vma);
void          nv_kern_vma_release(struct vm_area_struct *vma);

int           nv_kern_open(struct inode *, struct file *);
int           nv_kern_close(struct inode *, struct file *);
int           nv_kern_mmap(struct file *, struct vm_area_struct *);
unsigned int  nv_kern_poll(struct file *, poll_table *);
int           nv_kern_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
void          nv_kern_isr(int, void *, struct pt_regs *);
#ifdef CONFIG_PM_NOT_YET
int           nv_kern_pm(struct pm_dev *dev, pm_request_t rqst, void *data);
#endif

int           nv_kern_ctl_open(struct inode *, struct file *);
int           nv_kern_ctl_close(struct inode *, struct file *);

#ifdef CONFIG_PROC_FS
int           nv_kern_proc_read(char *, char **, off_t, int, int *, void *);
#endif

/***
 *** see nv.h for functions exported to other parts of resman
 ***/


/* character driver entry points */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
static struct file_operations nv_fops = {
    NULL,                   // lseek
    NULL,                   // read,
    NULL,                   // write
    NULL,                   // readdir
    nv_kern_poll,
    nv_kern_ioctl,
    nv_kern_mmap,
    nv_kern_open,
    NULL,                   // flush
    nv_kern_close,          // release
    NULL,                   // fsync
    NULL,                   // fasync
    NULL,                   // check_media_change
    NULL,                   // revalidate
    NULL,                   // lock
};
#else
static struct file_operations nv_fops = {
    owner:          THIS_MODULE,
    poll:           nv_kern_poll,
    ioctl:          nv_kern_ioctl,
    mmap:           nv_kern_mmap, 
    open:           nv_kern_open, 
    release:        nv_kern_close,
};
#endif

// Our reserved major device number.
// Set this to 0 to request dynamic major number assignment
int nv_major = NV_MAJOR_DEVICE_NUMBER;


// do not call this directly
void NV_iMSG(
    nv_state_t *nv,
    const char *kernel_message_level,
    const char *printf_format,
    va_list arglist
  )
{
    char nv_error_string[256];
    char *p;

    p = nv_error_string;
    
    if (kernel_message_level)
        p += sprintf(p, "%s", kernel_message_level);

    if (nv)
        if (nv->device_number == NV_CONTROL_DEVICE_NUMBER)
            p += sprintf(p, "NVctl: ");
        else
            p += sprintf(p, "NV%d: ", nv->device_number);
    else
        p += sprintf(p, "NV: ");

    p += vsprintf(p, printf_format, arglist);
    printk("%s\n", nv_error_string);
}

// should only be called using NV_DMSG macro
void NV_iDMSG(
    nv_state_t *nv,
    const char *printf_format,
    ...
)
{
    va_list arglist;

    va_start(arglist, printf_format);
    NV_iMSG(nv, KERN_INFO, printf_format, arglist);
    va_end(arglist);
}

void NV_EMSG(
    nv_state_t *nv,
    const char *printf_format,
    ...
  )
{
    va_list arglist;

    va_start(arglist, printf_format);
    NV_iMSG(nv, KERN_ERR, printf_format, arglist);
    va_end(arglist);
}


//
// The following are the supported pci vendor_id/device_id pairs
//

#define NVIDIA_VENDOR 0x10DE

static unsigned long pciIds[] = {
    (0x20 << 16) + NVIDIA_VENDOR,
    (0x21 << 16) + NVIDIA_VENDOR,
    (0x22 << 16) + NVIDIA_VENDOR,
    (0x23 << 16) + NVIDIA_VENDOR,
    (0x24 << 16) + NVIDIA_VENDOR,
    (0x25 << 16) + NVIDIA_VENDOR,
    (0x26 << 16) + NVIDIA_VENDOR,
    (0x27 << 16) + NVIDIA_VENDOR,
    (0x28 << 16) + NVIDIA_VENDOR,
    (0x29 << 16) + NVIDIA_VENDOR,
    (0x2A << 16) + NVIDIA_VENDOR,
    (0x2B << 16) + NVIDIA_VENDOR,
    (0x2C << 16) + NVIDIA_VENDOR,
    (0x2D << 16) + NVIDIA_VENDOR,
    (0x2E << 16) + NVIDIA_VENDOR,
    (0x2F << 16) + NVIDIA_VENDOR,

    (0xA0 << 16) + NVIDIA_VENDOR,

    (0x100 << 16) + NVIDIA_VENDOR,
    (0x101 << 16) + NVIDIA_VENDOR,
    (0x102 << 16) + NVIDIA_VENDOR,
    (0x103 << 16) + NVIDIA_VENDOR,

    (0x110 << 16) + NVIDIA_VENDOR,		
    (0x111 << 16) + NVIDIA_VENDOR,		
    (0x112 << 16) + NVIDIA_VENDOR,		
    (0x113 << 16) + NVIDIA_VENDOR,		

    (0x150 << 16) + NVIDIA_VENDOR,
    (0x151 << 16) + NVIDIA_VENDOR,
    (0x152 << 16) + NVIDIA_VENDOR,
    (0x153 << 16) + NVIDIA_VENDOR,

    (0x200<< 16) + NVIDIA_VENDOR,
    0
};


/***
 *** STATIC functions
 ***/

/*
** Returns 1 if device is ours, 0 otherwise.
*/

static int nvos_is_nv_device(
    unsigned char bus,
    unsigned char func,
    unsigned short vendor,
    unsigned short device
)
{
    unsigned long pciId = (device << 16) | vendor;
    unsigned long *id;

    NV_DMSG((nv_state_t *) 0,
            "is_nv_device: bus 0x%x fun 0x%x vendor 0x%x device 0x%x",
            bus, func, vendor, device);

    for (id = pciIds; *id != 0; id++)
        if (*id == pciId)
            return 1;

    return 0;
}

/*
** set which card is primary
*/
static int nvos_set_primary_card(nv_ioctl_primary_card_t *info)
{
    nv_linux_devices[info->index].nv_state.flags |= NV_FLAG_POSTED;

    return 0;
}

void nvos_probe_devices(void)
{
    int probenum = 0;
    unsigned char bus, func;
    int dev_count;
    nv_state_t *nv;
    nv_linux_state_t *nvl;

    NV_DMSG((nv_state_t *) 0, "nvos_probe_devices");

    //
    // Find nv devices and setup initial state.
    //
    dev_count = 0;
    while (PCIBIOS_SUCCESSFUL ==
           pcibios_find_class(PCI_CLASS_DISPLAY_VGA << 8, probenum,
                              &bus, &func))
    {
        unsigned short vendor_id, device_id;

        pcibios_read_config_word(bus, func, PCI_VENDOR_ID, &vendor_id);
        pcibios_read_config_word(bus, func, PCI_DEVICE_ID, &device_id);

        if (nvos_is_nv_device(bus, func, vendor_id, device_id))
        {
            unsigned short command_word;

            // Initialize bus-dependent config state.
            nvl = &nv_linux_devices[dev_count];
            nv = NV_STATE_PTR(nvl);
            nv->os_state = (void *) nvl;

            nv->bus = bus;
            nv->slot = func >> 3;
            nv->vendor_id = vendor_id;
            nv->device_id = device_id;
            nv->device_number = dev_count;

            // is the device enabled currently?
            pcibios_read_config_word(bus, func, PCI_COMMAND, &command_word);
            if ( ! (command_word & PCI_COMMAND_MEMORY) ||
                 ! (command_word & PCI_COMMAND_MASTER))
            {
                NV_DMSG(nv, "device not PCI enabled; enabling");
                command_word |= PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER;
                pcibios_write_config_word(bus, func, PCI_COMMAND, command_word);
            }

#ifdef KERNEL_2_4
            nv->reg_physical_address = pci_find_slot(bus, func)->resource[0].start;
            nv->reg_physical_address &= PCI_BASE_ADDRESS_IO_MASK;

            nv->fb_physical_address = pci_find_slot(bus, func)->resource[1].start;
            nv->fb_physical_address &= PCI_BASE_ADDRESS_MEM_MASK;
            
            nv->interrupt_line = pci_find_slot(bus, func)->irq;
            /* XXX not sure of correct way to read interrupt_pin & 
             * we don't appear to use it anyways...
             */
#else
#if LINUX_VERSION_CODE >= 0x20155
            nv->reg_physical_address = pci_find_slot(bus, func)->base_address[0];
            nv->reg_physical_address &= PCI_BASE_ADDRESS_IO_MASK;

            nv->fb_physical_address = pci_find_slot(bus, func)->base_address[1];
            nv->fb_physical_address &= PCI_BASE_ADDRESS_MEM_MASK;

            nv->interrupt_line = pci_find_slot(bus, func)->irq;
            /* XXX not sure of correct way to read interrupt_pin & 
             * we don't appear to use it anyways...
             */
#else
            pcibios_read_config_dword(bus, func, PCI_BASE_ADDRESS_0, &addr);
            nv->reg_physical_address = addr & PCI_BASE_ADDRESS_IO_MASK;

            pcibios_read_config_dword(bus, func, PCI_BASE_ADDRESS_1, &addr);
            nv->fb_physical_address = addr & PCI_BASE_ADDRESS_MEM_MASK;

            pcibios_read_config_byte(bus, func, PCI_INTERRUPT_LINE, &intr);
            nv->interrupt_line = intr;

            pcibios_read_config_byte(bus, func, PCI_INTERRUPT_PIN, &intr);
            nv->interrupt_pin = intr;
#endif
#endif

            /* check common error condition */
            if (nv->interrupt_line == 0)
            {
                printk("NV: I can't find an interrupt for your nvidia card\n");
                printk("NV: you may need to change your bios settings:\n");
                printk("NV: plug and play OS should be set to 'NO'\n");
                printk("NV: VGA IRQ should be set to 'YES'\n");
            }

            NV_DMSG(nv,
                    "bus: %d,  slot: %d,  devid: 0x%x, reg phys: %08x,  fb phys: %08x",
                    nv->bus, nv->slot, nv->device_id,
                    (unsigned int) nv->reg_physical_address,
                    (unsigned int) nv->fb_physical_address);

            if (++dev_count == NV_MAX_DEVICES)
            {
                NV_EMSG(nv, "maximum devices (%d) exceeded!", NV_MAX_DEVICES);
                break;
            }
        }
        probenum++;
    }

    NV_DMSG((nv_state_t *) 0,
            "nvos_probe_devices: found %d device%s", dev_count,
            dev_count == 1 ? "" : "s");
}


/* memory conversion routines */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 12)
static inline unsigned long 
nvos_get_kvirt_address(unsigned long address)
{
    pgd_t *pg_dir;
    pmd_t *pg_mid_dir;
    pte_t *pg_table;
    unsigned long ret_addr = 0;
    
    pg_dir = pgd_offset(current->mm, address);
    if (pgd_none(*pg_dir))
        goto done;

    pg_mid_dir = pmd_offset(pg_dir, address);
    if (pmd_none(*pg_mid_dir))
        goto done;

    pg_table = pte_offset(pg_mid_dir, address);
    if (!pte_present(*pg_table))
        goto done;

    ret_addr = pte_page(*pg_table) | NV_MASK_OFFSET(address);

  done:
    return ret_addr;
}

#else

static inline unsigned long 
nvos_get_kvirt_address(unsigned long address)
{
    pgd_t *pg_dir;
    pmd_t *pg_mid_dir;
    pte_t *pg_table;
    unsigned long ret_addr = 0;

    if (address > VMALLOC_START)
       pg_dir = pgd_offset_k(address);
    else
       pg_dir = pgd_offset(current->mm, address);

    if (pgd_none(*pg_dir))
        goto failed;

    pg_mid_dir = pmd_offset(pg_dir, address);
    if (pmd_none(*pg_mid_dir))
        goto failed;

    pg_table = pte_offset(pg_mid_dir, address);
    if (!pte_present(*pg_table))
        goto failed;

#if !defined(KERNEL_2_4)
    ret_addr = ((unsigned long)pte_page(*pg_table) | NV_MASK_OFFSET(address));
#elif !defined(IA64)
    if (ret_addr == 0)
    {
        ret_addr = ((unsigned long) __va(pte_val(*pg_table) & PAGE_MASK));
        ret_addr = (ret_addr | NV_MASK_OFFSET(address));
    }
#else
    if (ret_addr == 0)
    {
        ret_addr = ((unsigned long) __va(pte_val(*pg_table) & PAGE_MASK));
        ret_addr = (ret_addr | NV_MASK_OFFSET(address)) & 0xffffffff; 
    }
    else
    {
	/* If we're here, we have a system page (rather than a hw page).
           Earlier, we aligned the address to a hw page size (4096), but that 
           may be wrong. align the address to an os page size, so that adding 
           the page offset will be correct. 
         */
        ret_addr = ((unsigned long) page_address(pte_page(*pg_table)));
        ret_addr = (ret_addr | NV_MASK_OFFSET( (address & PAGE_MASK) ));
    }
#endif

    goto done;

  failed:
#if defined(IA64)
    /* on IA64, ioremappings are direct-mapped.. so give this a shot.. */
    if (address > VMALLOC_START)
        ret_addr = __pa(address);
#endif
    
  done:
    return ret_addr;
}
#endif

/* nvos_malloc explanation:
   we used to use vmalloc to allocate memory for dma buffers, etc.. but the 
   problem with that is that in systems with large amounts of physical memory 
   (1 gig, etc), it's easy to exhaust the kernel's virtual memory address space,
   while plenty of physical memory is still available. for that reason, we 
   directly manage physical pages themselves. we shouldn't have problems 
   exhausting memory until the system really is out of physical memory. 

   a second problem I've encountered is problems with fragmenting "contiguous 
   memory blocks". I originally tried to allocate as many contiguous pages as 
   I could for my allocations, but that eventually left memory fragmented, 
   and only single pages free. Since the fork system call needs 2 contiguous 
   pages to succeed, we would eventually keep fork from working, even though 
   there were massive amounts of single pages free.

   so now, we allocate single physical pages at a time, until we have all we 
   need. this seems to work pretty well, and minimize system fragmentation.
 */
static void *
nvos_malloc(unsigned long size)
{
    unsigned int pages_needed = (size >> PAGE_SHIFT);
    unsigned long *page_list = NULL, *page_ptr = NULL;

    /* allocate a pointer for each page, plus an int for the num pages  */
    NV_VMALLOC(page_list, sizeof(unsigned long *) * (pages_needed + 2));
    if (page_list == NULL)
        return NULL;

    *page_list++ = (unsigned long) pages_needed; /* number of pages allocated */
    page_ptr = page_list;                        /* starts actual page list */

#if defined(IA64)
    {
        int i;
        void *tmpaddr;

        NV_VMALLOC(tmpaddr, size);
        if (tmpaddr == 0x0)
            goto failed;

        /* save the vmalloc address for vfree.. */
        *page_list++ = (unsigned long) tmpaddr;
        page_ptr = page_list;

        for (i = 0; i < pages_needed; i++)
        {
            *page_ptr = nv_get_phys_address( (unsigned long) tmpaddr);

            /* lock the page for dma purposes */
            mem_map_reserve(GET_MAP_NR(*(page_ptr)));

            tmpaddr += PAGE_SIZE;
            page_ptr++;
        }
    }
#else
    {
        NVRM_LOCK_SETUP();
        NVRM_LOCK_RELEASE();
        while (pages_needed)
        {
            unsigned long addr = __get_free_pages(GFP_KERNEL, 0);
            if (addr == 0x0)
                goto failed;

            /* lock the page for dma purposes */
            mem_map_reserve(GET_MAP_NR(__pa(addr)));

            *page_ptr++ = __pa(addr);
            pages_needed--;
        }
        NVRM_LOCK_GRAB();
    }
#endif

    return (void*) page_list;

#if !defined(IA64)
failed:
    /* if we've run out of memory, clean up after ourselves */
    printk("NVRM: BAD NEWS! We've run out of kernel memory and failed our allocations!\n");
    {
        unsigned long *page_del = page_list;
        while (page_del != page_ptr)
        {
            mem_map_unreserve(GET_MAP_NR(*page_del));
            free_page( (unsigned long) __va(*page_del));
            page_del++;
        }
        page_list -= 1;   /* back up over num pages */
        NV_VFREE(page_list);
    }
    return NULL;
#endif
}

static void
nvos_free(void **page_list)
{
    int i, size = *(unsigned int *)(page_list - 1);
    unsigned long *page_ptr = (unsigned long *) page_list;
    for (i = 0; i < size; i++)
    {
        mem_map_unreserve(GET_MAP_NR(*page_ptr));
#if !defined(IA64)
        free_page( (unsigned long) __va(*page_ptr++));
#endif
    }

    /* back up over num pages, so we free the correct address */
    page_list -= 1;

#if defined(IA64)
    /* is this a typo?? */
    NV_VFREE(*page_list);
#else
    NV_VFREE(page_list);
#endif
}

#ifdef CONFIG_PROC_FS
static void 
nvos_proc_create(void)
{
    nv_linux_state_t *tnvl;
    struct proc_dir_entry *p;
    char name[6];
    int i = 0;
      
    nv_proc_entry = create_proc_entry("nv", S_IFDIR|S_IRUGO|S_IXUGO, &proc_root);
	
    if (nv_proc_entry == NULL) {
        printk("NVRM: unable to create /proc/nv\n");
        return;
    }
#ifndef KERNEL_2_2	
    nv_proc_entry->owner = THIS_MODULE;
#endif

    for (tnvl = nv_linux_devices; tnvl < nv_linux_devices + NV_MAX_DEVICES; tnvl++, i++)
    {
        if ((NV_STATE_PTR(tnvl))->device_id)
        {
            sprintf(name, "card%d", i);
            p = create_proc_entry(name, S_IFREG|S_IRUGO, nv_proc_entry);
            p->data = (void*) tnvl;
            p->read_proc = nv_kern_proc_read;
#ifndef KERNEL_2_2
            p->owner = THIS_MODULE;
#endif
        }
    }
}

static void 
nvos_proc_remove(void)
{
    if (nv_proc_entry != NULL) remove_proc_entry("nv", &proc_root);
}

#endif /* CONFIG_PROC_FS */


/*
 * Given a virtual address, fid the 'at' that owns it
 * Uses the physical address as the key.
 */
static nv_alloc_t *nvl_find_alloc(
    nv_linux_state_t    *nvl,
    unsigned long  address,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;
    void *physaddr, *offset;
    unsigned int i;

    address = address & PAGE_MASK;
    physaddr = (void *) nv_get_phys_address(address);

    if (physaddr == NULL)
        goto done;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nvl->alloc_queue; at; *prev = at, at = at->next)
    {
        if (at->page_table)
        {
            /* alloc/free pages will be sending the page_table address */
            if (address == ((unsigned long) at->page_table & PAGE_MASK))
                return at;

            for (i = 0; i < at->num_pages; i++)
            {
                offset = at->page_table[i];
                if ( (physaddr >= offset) && 
                     (physaddr < offset + PAGE_SIZE - 1) )
                    return at;
            }
        }
    }

done:
    /* failure is not necessarily an error if the caller 
       was just probing an address */
    NV_DMSG(NV_STATE_PTR(nvl), "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}

/*
 * Given a virtual address, fid the 'at' that owns it
 * Uses the physical address as the key.
 */
static nv_alloc_t *nvl_find_agp_alloc(
    nv_linux_state_t    *nvl,
    unsigned long  address,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nvl->alloc_queue; at; *prev = at, at = at->next)
    {
        if (NV_STATE_PTR(nvl)->agpgart)
        {
            if (at->agp_mapping &&
                ((address == (unsigned long)at->agp_mapping) || 
                (address == at->vma->vm_start))) {
                return at;
            }
        } else {
            if ((void *) address == (void *) at->page_table)
                return at;
        }
    }

    /* failure is not necessarily an error if the caller 
       was just probing an address */
    NV_DMSG(NV_STATE_PTR(nvl), "could not find map for vm 0x%lx", address);
    return (nv_alloc_t *) 0;
}


/*
 * We allocated system/agp memory in an ioctl before a mmap, and now we need to
 * retreive it. Find the allocation based on number of pages and having no vma
 * associated with the address.
 * XXX getting 2 allocations mixed up at this point shouldn't be a problem,
 * since the memory isn't associated with a client/user yet.
 */
static nv_alloc_t *nvl_find_alloc_by_num_pages(
    nv_linux_state_t    *nvl,
    unsigned long  num_pages,
    nv_alloc_t   **prev
)
{
    nv_alloc_t *dummy_prev;
    nv_alloc_t *at;

    if (prev == (nv_alloc_t **) 0)
        prev = &dummy_prev;

    *prev = (nv_alloc_t *) 0;
    for (at = nvl->alloc_queue; at; *prev = at, at = at->next)
    {
        if ((at->num_pages == num_pages) && 
            (at->process_id == current->pid) &&
            (at->vma == NULL) )
            return at;
    }

    /* failure is not necessarily an error if the caller 
       was just probing an address */
    NV_DMSG(NV_STATE_PTR(nvl), "could not find a mapping with %d pages", 
        num_pages);

    return (nv_alloc_t *) 0;
}


static void nv_interrupts_disable(
    nv_state_t *nv
)
{
    nv->regs[NV_PMC_INTR_EN_0 / 4] = 0;
}

static void nv_interrupts_enable(
    nv_state_t *nv
)
{
    nv->regs[NV_PMC_INTR_EN_0 / 4] = 1;
}

/* XXX: need to minimize mappings
        the pci config space only tracks how much space is reserved.
        It doesn't really know how big stuff is.
 */
static int nv_map_device(
    nv_state_t *nv
)
{
    unsigned long size;

    // map the register space
    nv->regs = ioremap(nv->reg_physical_address, 16 * 1024 * 1024);

    /* map the frame buffer
       XXX should map measured size of fb, not fixed size
           unfortunately I don't know how to get the fb size.
           We have to know what it is before we call initHal()...
           current hack is to just keep mapping until we get something we can use
        NOTE:
           2.2.12 kernels seem to have a bug where they return the physical
           address if they can't get the memory, hence the check against physaddr
     */

    size = 64 * 1024 * 1024;

    do {
        nv->fb = ioremap(nv->fb_physical_address, size);
        if ((nv->fb == (void *) 0) || (nv->fb == (void *)(NV_UINTPTR_T)nv->fb_physical_address))
        {
            nv->fb = 0;
            size /= 2;
        }
    } while ((nv->fb == (void *) 0) && (size >= (4 * 1024 * 1024)));

    if (nv->fb == 0)
        return -ENXIO;

    nv->regs_u = (nv_phwreg_t) nv->regs;
    nv->fb_u = (nv_phwreg_t) nv->fb;
    return 0;
}

static void nv_unmap_device(
    nv_state_t *nv
)
{
#if !defined(IA64)
    /* we're given the address of a pre-mapped mapping that we can't
     * free under IA64, so skip this.
     */
    if (nv->regs)
        NV_VFREE(nv->regs);
    if (nv->fb)
        NV_VFREE(nv->fb);
#endif
    nv->regs = nv->fb = (void *) 0;
    nv->regs_u = nv->fb_u = (void *) 0;
}


/***
 *** EXPORTS to Linux Kernel
 ***/

int init_module(void)
{
    nv_linux_state_t *nvl;
    int rc;

    printk("NVRM: loading NVIDIA kernel module version %d.%d-%d\n", 
        NV_MAJOR_VERSION, NV_MINOR_VERSION, NV_PATCHLEVEL);

    // Register driver
    rc = register_chrdev(nv_major, "nvidia", &nv_fops);
    if (rc < 0) {
        NV_EMSG((nv_state_t *) 0,
                "init_module: register failed %d", rc);
        goto failed;
    }
    // if we requested dynamic major assignment, then 'rc' will have new number
    if (nv_major == 0)
        nv_major = rc;

    NV_DMSG((nv_state_t *) 0, "init_module: major number %d", nv_major);

    osMemSet(&nv_linux_devices, 0, sizeof(nv_linux_state_t) * NV_MAX_DEVICES);

    nvos_probe_devices();

    // init all the bottom half structures
    for (nvl = nv_linux_devices; nvl < nv_linux_devices + NV_MAX_DEVICES; nvl++)
    {
        nvl->bh = &nv_bottom_halves[nvl - nv_linux_devices];
        nvl->bh->routine = RmIsrBottomHalf;
        nvl->bh->data = (void *) 0;
        nvl->bh->sync = 0;
    }

#ifdef CONFIG_PM_NOT_YET
    /* XXX PM egads, is this the right place to do this? */
    printk("registering Power Management\n");
    pm_nv_dev = pm_register(PM_PCI_DEV, PM_SYS_VGA, nv_kern_pm);
#endif

    // Init the resource manager
    if ( ! RmInitRm())
    {
        NV_EMSG((nv_state_t *) 0, "RmInitRm() failed");
        rc = -EIO;
        goto failed;
    }

#ifdef CONFIG_PROC_FS
    nvos_proc_create();
#endif

    nv_lock_init();

    return 0;

 failed:
    if (nv_major > 0)
        unregister_chrdev(nv_major, "nvidia");
    return rc;
}

void cleanup_module(void)
{
    int rc;
    nv_linux_state_t *nvl;
    
#ifdef CONFIG_PROC_FS
    nvos_proc_remove();
#endif

    NV_DMSG((nv_state_t *) 0, "cleanup_module");

#ifdef CONFIG_PM_NOT_YET
    /* XXX PM egads, is this the right place to do this? */
    printk("unregistering Power Management\n");
    pm_unregister(pm_nv_dev);
#endif

    // Shutdown the resource manager
    RmShutdownRm();

    // make sure we freed up all the mappings
    // kernel should do this automagically before calling close
    for (nvl = &nv_linux_devices[0]; nvl < nv_linux_devices + NV_MAX_DEVICES; nvl++)
        if (nvl->alloc_queue)
            NV_EMSG(NV_STATE_PTR(nvl),
                    "still have vm que at cleanup_module(): 0x%x to 0x%x",
                    nvl->alloc_queue->vma->vm_start,
                    nvl->alloc_queue->vma->vm_end);

    if ((rc = unregister_chrdev(nv_major, "nvidia")) < 0)
        NV_EMSG((nv_state_t *) 0,
                "cleanup_module: unregister nv failed %d", rc);
}


/* this is only called when the vmas are duplicated.
 * this appears to only happen when the process is cloned to create
 * a new process, and not when the process is threaded. 
 *
 * increment the usage count for the physical pages, so when this
 * clone unmaps the mappings, the pages are not deallocated under
 * the original process.
 */ 
void 
nv_kern_vma_open(struct vm_area_struct *vma)
{
    MOD_INC_USE_COUNT;
}


void 
nv_kern_vma_release(struct vm_area_struct *vma)
{
    MOD_DEC_USE_COUNT;
}


struct vm_operations_struct nv_vm_ops = {
    nv_kern_vma_open,
    nv_kern_vma_release,  /* "close" */
};


/*
** nv_kern_open
**
** nv driver open entry point.  Sessions are created here.
*/
int nv_kern_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = (nv_state_t *) 0;
    nv_linux_state_t *nvl = (nv_linux_state_t *) 0;
    nv_file_private_t *nvfp;
    int devnum;
    int rc, status;
    
    /* Grab a file private area and save 'nv' in file structure */
    NV_KMALLOC(nvfp, sizeof(nv_file_private_t));
    if ( ! nvfp)
        return -ENOMEM;
    memset(nvfp, 0, sizeof(*nvfp));
    file->private_data = nvfp;

    /* for control device, just jump to its open routine */
    /* after setting up the private data */
    if (NV_DEVICE_IS_CONTROL_DEVICE(inode->i_rdev))
        return nv_kern_ctl_open(inode, file);

    /* what device are we talking about? */
    devnum = NV_DEVICE_NUMBER(inode->i_rdev);
    if (devnum >= NV_MAX_DEVICES)
    {
        rc = -ENODEV;
        goto failed;
    }

    MOD_INC_USE_COUNT;

    nvl = &nv_linux_devices[devnum];
    nv = NV_STATE_PTR(nvl);

    NV_DMSG(nv, "open by process %d (current is 0x%x)", current->pid, current);

    NV_HIDE_IN_FILEP(file, nvl);

    /*
     * map the memory and allocate isr on first open
     */

    if ( ! (nv->flags & NV_FLAG_OPEN))
    {
        if (nv->device_id == 0)
        {
            NV_EMSG((nv_state_t *) 0, "open of nonexistent device %d", devnum);
            rc = -ENXIO;
            goto failed;
        }

        if ((rc = nv_map_device(nv)))
        {
            NV_EMSG(nv, "nv_map_device failed");
            goto failed;
        }

        if ( ! RmInitAdapter(nv))
        {
            NV_EMSG(nv, "RmInitAdapter failed");
            rc = -EIO;
            goto failed;
        }
        nv->flags |= NV_FLAG_POSTED;

        status = request_irq(nv->interrupt_line, nv_kern_isr,
                             SA_INTERRUPT | SA_SHIRQ, "nvidia",
                             (void *) nvl);
        if (status != 0)
        {
            if ( nv->interrupt_line && (status == -EBUSY) )
            {
                printk("NV: Tried to get irq %d, but another driver", 
                    (unsigned int) nv->interrupt_line);
                printk(" has it and is not sharing it.\n");
                printk("NV: you may want to verify that an audio driver");
                printk(" isn't using the irq\n");
            }
            NV_EMSG(nv, "isr request failed 0x%x", status);
            rc = -EIO;
            goto failed;
        }

#ifdef KERNEL_2_4
        NV_VMALLOC(nvl->event_queue, sizeof(struct __wait_queue_head));
        osMemSet(nvl->event_queue, 0, sizeof(struct __wait_queue_head));
        if (nvl->event_queue == NULL)
            goto failed;

        init_waitqueue_head(GET_EVENT_QUEUE(nvl));
#else
        nvl->event_queue = NULL;
#endif

        nv->flags |= NV_FLAG_OPEN;
    }

    nv->usage_count++;

    return 0;

 failed:
    MOD_DEC_USE_COUNT;
    if (nv)
        nv_unmap_device(nv);
    return rc;
}


/*
** nv_kern_close
**
** Master driver close entry point.
*/

int nv_kern_close(
    struct inode *inode,
    struct file *file
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);
    unsigned long lock_flags;

    /* for control device, just jump to its open routine */
    /* after setting up the private data */
    if (NV_DEVICE_IS_CONTROL_DEVICE(inode->i_rdev))
        return nv_kern_ctl_close(inode, file);

    NV_DMSG(nv, "close");

    nv_lock_grab(&lock_flags);

    nv_unix_free_all_unused_clients(nv, current->pid, (void *) file);

    if (--nv->usage_count == 0)
    {
        int counter = 0;

        /* turn off interrupts.
        ** be careful to make sure any pending bottom half gets run
        **  or disabled before calling RmShutdownAdapter() since
        **  it will free up the pdev.  This is hard to see on single
        **  cpu systems, but easy on dual cpu :-)
        */
        nv_interrupts_disable(nv);
        
        /* give it a moment to allow any bottom half to run */
        /* have to release lock of course */

        /* drop the lock, so we can make sure all interrupts are done */
        /* being processed, without locking them out */
        nv_lock_release(lock_flags);

#define MAX_BH_TASKS 10
        while ((nv->bh_count) && (counter < MAX_BH_TASKS))
        {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ/50);
            counter++;
        }

        /* free the irq, which may block until any pending interrupts */
        /* are done being processed. */
        free_irq(nv->interrupt_line, (void *) nv);

        /* regain lock and free the irq */
        nv_lock_grab(&lock_flags);

        nvl->bh->data = (void *) 0;

        RmShutdownAdapter(nv);

        (void) nv_unmap_device(nv);

        /* make sure we have freed up all the mappings */
        /* kernel should do this automagically before calling close */
        if (nvl->alloc_queue)
            NV_EMSG(nv,
                    "still have vm que at nv_close(): 0x%x to 0x%x",
                    nvl->alloc_queue->vma->vm_start,
                    nvl->alloc_queue->vma->vm_end);

        /* leave INIT flag alone so we don't reinit every time */
        nv->flags &= ~(NV_FLAG_OPEN | NV_FLAG_WAITING);
    }

    /* free up our per file private data */
    if (file->private_data)
        NV_KFREE(file->private_data);
    file->private_data = (void *) 0;

    nv_lock_release(lock_flags);

    MOD_DEC_USE_COUNT;

    return 0;
}

int nv_kern_mmap(
    struct file  *file,
    struct vm_area_struct *vma
)
{
    int pages;
    nv_alloc_t *at;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);
    unsigned long lock_flags;

    NV_DMSG(nv, "mmap([0x%lx-0x%lx] off=0x%lx", 
        vma->vm_start, 
        vma->vm_end, 
        LINUX_VMA_OFFS(vma));

    // be a bit paranoid for now
    if ((NV_MASK_OFFSET(LINUX_VMA_OFFS(vma))) ||
        (NV_MASK_OFFSET(vma->vm_start)) ||
        (NV_MASK_OFFSET(vma->vm_end)))
    {
        return -ENXIO;
    }

    pages = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;

    // max is 64M
    if (pages > (64 * 1024 * 1024) / PAGE_SIZE)
        return -ENOMEM;

    // we have our own version to keep the module count right
    vma->vm_ops = &nv_vm_ops;
    
    /*
     * figure out the range and map it in
     */

    
    /* NV reg space */
    if ((LINUX_VMA_OFFS(vma) >= NV_MMAP_REG_OFFSET) &&
        ((LINUX_VMA_OFFS(vma) + (vma->vm_end - vma->vm_start)) <= NV_MMAP_FB_OFFSET))
    {
#if defined(IA64)
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
        if (remap_page_range(vma->vm_start,
                             (u32) (nv->reg_physical_address) + LINUX_VMA_OFFS(vma) - NV_MMAP_REG_OFFSET,
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        /* mark it as IO so that we don't dump it on core dump */
        vma->vm_flags |= VM_IO;
    }

    /* NV fb space */
    else if ((LINUX_VMA_OFFS(vma) >= NV_MMAP_FB_OFFSET) &&
             ((LINUX_VMA_OFFS(vma) + (vma->vm_end - vma->vm_start)) <= NV_MMAP_ALLOCATION_OFFSET))
    {
#if defined(IA64)
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
        if (remap_page_range(vma->vm_start,
                             (u32) (nv->fb_physical_address) + LINUX_VMA_OFFS(vma) - NV_MMAP_FB_OFFSET,
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        // mark it as IO so that we don't dump it on core dump
        vma->vm_flags |= VM_IO;
    }

    /* Magic allocator */
    else if (LINUX_VMA_OFFS(vma) == NV_MMAP_ALLOCATION_OFFSET)
    {
        unsigned long page = 0, pos, start;
        int i = 0;

        // protect nvl->alloc_queue with global RM lock

        nv_lock_grab(&lock_flags);
        at = nvl_find_alloc_by_num_pages(nvl, pages, NULL);
        nv_lock_release(lock_flags);

        if (at == NULL)
        {
            printk("NVRM: couldn't find pre-allocated memory!\n");
            return -EAGAIN;
        }

        at->vma = vma;
        start = vma->vm_start;
        while (pages--)
        {
            page = (unsigned long) at->page_table[i++];
            if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
              	return -EAGAIN;
            start += PAGE_SIZE;
            pos += PAGE_SIZE;
       	}

        /* prevent the swapper from swapping it out */
        vma->vm_flags |= VM_LOCKED;
    }

    /* AGP allocator */
    else if (LINUX_VMA_OFFS(vma) == NV_MMAP_AGP_OFFSET)
    {
        // protect nvl->alloc_queue with global RM lock

        nv_lock_grab(&lock_flags);
        at = nvl_find_alloc_by_num_pages(nvl, pages, NULL);
        nv_lock_release(lock_flags);

        if (at == NULL)
        {
            printk("NVRM: couldn't find pre-allocated memory!\n");
            return -EAGAIN;
        }

        at->vma = vma;

        if (nv->agpgart)
            KernMapAGPPages(vma, at->priv_data);
        else
        {
            // Protect structures modified by NvMapAGPPages

            nv_lock_grab(&lock_flags);

            NvMapAGPPages(nv->pdev, (void **) &vma->vm_start, 
                          at->class, at->priv_data);

            nv_lock_release(lock_flags);

            /* assumes we're not supporting IA64 AGP ourselves */
            at->agp_mapping = (void *)nv_get_phys_address(vma->vm_start);
        }
        /* prevent the swapper from swapping it out */
        vma->vm_flags |= VM_LOCKED;
    } else {
        printk("NVRM: unknown memory type\n");
        return -EINVAL;
    }

    vma->vm_file = file;

    /* just increment usage count, rather than calling vma_open */
    MOD_INC_USE_COUNT;

    return 0;
}


unsigned int nv_kern_poll(
    struct file *file,
    poll_table *wait
)
{
    unsigned int mask = 0;
    nv_linux_state_t *nvl;
    nv_file_private_t *nvfp;
    unsigned long lock_flags;

    nvfp = file->private_data;
    nvl = NV_GET_NVL_FROM_FILEP(file);

    NV_DMSG(NV_STATE_PTR(nvl), "poll()");

    // add us to the list
    poll_wait(file, GET_EVENT_QUEUE(nvl), wait);

    nv_lock_grab(&lock_flags);

    // wake the user on any file-specific event, or a general vblank
    if (nvfp->any_fired_notifiers || nvl->vblank_notifier)
    {
        if (nvl->vblank_notifier)
        {
            // don't clear vblank_notifier until everyone's been notified
            nvl->waiting_for_vblank--;
            if (nvl->waiting_for_vblank == 0)
            {
                nvl->vblank_notifier = 0;
                NV_STATE_PTR(nvl)->flags &= ~NV_FLAG_WAITING;
            }
        }

        // tell user some have been posted
        mask |= POLLPRI;
        NV_DMSG(NV_STATE_PTR(nvl), "poll() live on entry!");
    } else {
        // keep track of how many people are waiting for vblank
        nvl->waiting_for_vblank++;
        NV_STATE_PTR(nvl)->flags &= NV_FLAG_WAITING;
    }

    nv_lock_release(lock_flags);

    return mask;
}

//
// nv_kern_ioctl
//
// nv driver ioctl entry point.
//

#ifdef __KERNEL__
#define COPYIN(dst, src, bytes)     copy_from_user((dst), (src), (bytes))
#define COPYOUT(dst, src, bytes)   copy_to_user((dst), (src), (bytes))
#else
#define COPYIN(dst, src, bytes)     memcpy((dst), (src), (bytes)), 0
#define COPYOUT(dst, src, bytes)   memcpy((dst), (src), (bytes)), 0
#endif

#define DO_OR_DIE(n) { if (n) { status = -EFAULT; goto done; } }

/*
 * some ioctl's can only be done on actual device, others only on the control device
 */
#define CTL_DEVICE_ONLY(nv) { if ( ! ((nv)->flags & NV_FLAG_CONTROL)) { status = -EINVAL; goto done; } }

#define ACTUAL_DEVICE_ONLY(nv) { if ((nv)->flags & NV_FLAG_CONTROL) { status = -EINVAL; goto done; } }


/* todo:
   need ioctl to raise a thread priority that is not superuser
       set its priority to SCHED_FIFO which is simple
       priority scheduling w/ disabled timeslicing
*/

int nv_kern_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int cmd,
    unsigned long i_arg)
{
    int size;
    int status = 0;
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    void *arg = (void *) i_arg;

    nvl = NV_GET_NVL_FROM_FILEP(file);
    nv = NV_STATE_PTR(nvl);

    NV_DMSG(nv, "ioctl(%d, 0x%x)", _IOC_NR(cmd), (unsigned int) i_arg);

    size = _IOC_SIZE(cmd);

    switch (cmd)
    {
        /* debug tool; zap the module use count so we can unload driver */
        /*             even if it is confused */
        case NV_IOCTL_MODULE_RESET:
            atomic_set(&__this_module.uc.usecount, 1);
            break;

        /* enable/disable interrupts 
           used by X driver to make card quiet while Xserver unnecessarily 
           diddles around in its PCI config space 
         */
        case NV_IOCTL_INTERRUPTS:
        {
            int state = i_arg;
            ACTUAL_DEVICE_ONLY(nv);
            NV_DMSG(nv, "%sabling interrupts", state ? "En" : "Dis");
            if (state)
                nv_interrupts_enable(nv);
            else
                nv_interrupts_disable(nv);
            break;
        }

        case NV_IOCTL_MEMORY_VTOP:
        {
            nv_ioctl_memory_vtop_t params;
            nv_ioctl_memory_vtop_t *userbuf = arg;

            ACTUAL_DEVICE_ONLY(nv);

            /* copy in the user command buffer */
            DO_OR_DIE(COPYIN(&params, userbuf, sizeof(params)));
            
            status = nv_vtop(&params, 1);

            break;
        }

        /* pass out info about the card */
        case NV_IOCTL_CARD_INFO:
        {
            nv_ioctl_card_info_t ci;
            nv_ioctl_card_info_t *userbuf = arg;
            nv_linux_state_t *tnvl;
            int i;

            CTL_DEVICE_ONLY(nv);

            for (i = 0, tnvl = nv_linux_devices; tnvl < nv_linux_devices + NV_MAX_DEVICES; tnvl++, i++)
            {
                nv_state_t *tnv = NV_STATE_PTR(tnvl);
                (void) memset(&ci, 0, sizeof(ci));
                if (tnv->device_id)
                {
                    ci.flags = NV_IOCTL_CARD_INFO_FLAG_PRESENT;
                    ci.bus = tnv->bus;
                    ci.slot = tnv->slot;
                    ci.vendor_id = tnv->vendor_id;
                    ci.device_id = tnv->device_id;
                    ci.interrupt_line = tnv->interrupt_line;
                }
                /* copy the card info out to user */
                /* just copying 0's for nonexistent cards */
                DO_OR_DIE(COPYOUT(userbuf + i, &ci, sizeof(ci)));
            }
            break;
        }

	/* set a card to be primary (not post it) */
	case NV_IOCTL_PRIMARY_CARD:
	{
	    nv_ioctl_primary_card_t params;
	    nv_ioctl_primary_card_t *userbuf = arg;

	    CTL_DEVICE_ONLY(nv);

	    /* copy in the user command buffer */
	    DO_OR_DIE(COPYIN(&params, userbuf, sizeof(params)));

	    status = nvos_set_primary_card(&params);

	    break;
	}

        default:
            status = nvExtEscape(file, nv, cmd, arg) ? 0 : -EINVAL;
            break;
    }

 done:
    return status;
}

/*
 * driver receives an interrupt
 *    if someone waiting, then hand it off.
 */
void nv_kern_isr(
    int   irq,
    void *arg,
    struct pt_regs *regs
)
{
    nv_linux_state_t *nvl = (void *) arg;
    nv_state_t *nv = NV_STATE_PTR(nvl);
    U032 need_to_run_bottom_half = 0;
    ULONG lock_flags;

    // printk(".");
    RmIsr(nv->device_number, &need_to_run_bottom_half);
    if (need_to_run_bottom_half)
    {
    nv_lock_grab(&lock_flags);
        nv->bh_count++;
        nvl->bh->data = nv->pdev;
        queue_task(nvl->bh, &tq_immediate);
        mark_bh(IMMEDIATE_BH);
    nv_lock_release(lock_flags);
    }
}

#ifdef CONFIG_PM_NOT_YET
/* kernel calls us with a power management event */
int
nv_kern_pm(
    struct pm_dev *dev, 
    pm_request_t rqst, 
    void *data
)
{
    /* XXX PM HACK!! for now, let's try this */
    nv_linux_state_t *lnv = &nv_linux_devices[0];

    printk("nv_kern_pm event: rqst 0x%x data 0x%lx\n", 
        (unsigned int) rqst, 
        (unsigned long) data);

    switch (rqst)
    {
        case PM_RESUME:
            printk("NVRM: received PM resume event\n");
            RmPowerManagement(NV_STATE_PTR(lnv)->pdev, 0, rqst);
            break;

        case PM_SUSPEND:
            printk("NVRM: received PM suspend event\n");
            RmPowerManagement(NV_STATE_PTR(lnv)->pdev, 0, rqst);
            break;

        default:
            printk("NVRM: received unknown PM event: 0x%x\n", rqst);
            return 1;
    }
    return 0;
}
#endif

/*
** nv_ctl_open
**
** nv control driver open entry point.  Sessions are created here.
*/
int nv_kern_ctl_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = (nv_state_t *) 0;

    nv = (nv_state_t *) &nv_ctl_device;
    nv->device_number = NV_CONTROL_DEVICE_NUMBER;

    /* save the nv away in file->private_data */
    NV_HIDE_IN_FILEP(file, nv);

    MOD_INC_USE_COUNT;

    NV_DMSG(nv, "open");

    nv->flags |= NV_FLAG_OPEN + NV_FLAG_CONTROL;

    nv->usage_count++;

    return 0;
}


/*
** nv_ctl_close
*/
int nv_kern_ctl_close(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = NV_GET_NV_FROM_FILEP(file);
    unsigned long lock_flags;

    NV_DMSG(nv, "close");

    nv_lock_grab(&lock_flags);
    if (--nv->usage_count == 0)
        nv->flags = 0;

    nv_unix_free_all_unused_clients(nv, current->pid, (void *) file);

    /* free up our per file private data */
    if (file->private_data)
        NV_KFREE(file->private_data);
    file->private_data = (void *) 0;
    nv_lock_release(lock_flags);

    MOD_DEC_USE_COUNT;

    return 0;
}

#ifdef CONFIG_PROC_FS

#ifndef PCI_DEVICE_ID_INTEL_82443GX_0
  #define PCI_DEVICE_ID_INTEL_82443GX_0    0x71a0
#endif
#ifndef PCI_DEVICE_ID_INTEL_810_0
  #define PCI_DEVICE_ID_INTEL_810_0        0x7120
#endif
#ifndef PCI_DEVICE_ID_INTEL_810_E_0
  #define PCI_DEVICE_ID_INTEL_810_E_0      0x7124
#endif
#ifndef PCI_DEVICE_ID_INTEL_840_0
  #define PCI_DEVICE_ID_INTEL_840_0        0x1a21
#endif
#ifndef PCI_DEVICE_ID_AL_M1541_0
  #define PCI_DEVICE_ID_AL_M1541_0         0x1541
#endif
#ifndef PCI_DEVICE_ID_AMD_IRONGATE_0
  #define PCI_DEVICE_ID_AMD_IRONGATE_0     0x7006
#endif
#ifndef PCI_DEVICE_ID_VIA_8501_0
  #define PCI_DEVICE_ID_VIA_8501_0         0x0501
#endif
#ifndef PCI_DEVICE_ID_VIA_8371_0
  #define PCI_DEVICE_ID_VIA_8371_0         0x0391
#endif
#ifndef PCI_DEVICE_ID_VIA_8363_0
  #define PCI_DEVICE_ID_VIA_8363_0         0x0305
#endif
#ifndef PCI_DEVICE_ID_VIA_82C691_0
  #define PCI_DEVICE_ID_VIA_82C691_0       0x0691
#endif

#define PCI_DEVICE_ID_RIVA_TNT1            0x0020
#define PCI_DEVICE_ID_RIVA_TNT2            0x0028
#define PCI_DEVICE_ID_RIVA_TNT2U           0x0029
#define PCI_DEVICE_ID_VANTA                0x002c
#define PCI_DEVICE_ID_ULTRA                0x002d
#define PCI_DEVICE_ID_INTEGRATED_TNT2      0x00a0
#define PCI_DEVICE_ID_GEFORCE              0x0100
#define PCI_DEVICE_ID_GEFORCEDDR           0x0101
#define PCI_DEVICE_ID_QUADRO               0x0103
#define PCI_DEVICE_ID_GEFORCE2             0x0110
#define PCI_DEVICE_ID_GEFORCE2_1           0x0111
#define PCI_DEVICE_ID_GEFORCE2_3           0x0113
#define PCI_DEVICE_ID_GEFORCE2GTS          0x0150
#define PCI_DEVICE_ID_GEFORCE2GTSDDR       0x0151
#define PCI_DEVICE_ID_GEFORCE2GTS_2        0x0152
#define PCI_DEVICE_ID_QUADRO2              0x0153
#define PCI_DEVICE_ID_NVIDIA_CARD0         0x0200
#define PCI_DEVICE_ID_NVIDIA_CARD1         0x0201
#define PCI_DEVICE_ID_NVIDIA_CARD2         0x0202
#define PCI_DEVICE_ID_NVIDIA_CARD3         0x0203

int nv_kern_proc_read(
    char *page,
    char **start,
    off_t off,
    int count,
    int *eof,
    void *data)
{
    struct pci_dev *dev;
    nv_state_t *tnv;
    unsigned char bus, func, cap_ptr = 0x00;
    unsigned int cap_id, status, command;
    unsigned short device_id, vendor_id;
    char chipset[25], device[25];
    int len = 0;

    tnv = NV_STATE_PTR((nv_linux_state_t*) data);

    *start = 0;
    *eof   = 1;

    len += sprintf(page+len, "----- Driver Info ----- \n");
    len += sprintf(page+len, "NVRM Version: %d.%d-%d\n", 
        NV_MAJOR_VERSION, NV_MINOR_VERSION, NV_PATCHLEVEL);
    
#ifdef KERNEL_2_2
    for (dev = pci_devices; dev; dev = dev->next) {
#else    
    pci_for_each_dev(dev) {
#endif
        if ( (dev->device == tnv->device_id) &&
             (dev->vendor == tnv->vendor_id) ) 
        {
            switch (dev->device) {
                case PCI_DEVICE_ID_RIVA_TNT1:
                    sprintf(device, "Riva TNT1"); break;
                case PCI_DEVICE_ID_RIVA_TNT2:
                    sprintf(device, "Riva TNT2"); break;
                case PCI_DEVICE_ID_RIVA_TNT2U:
                    sprintf(device, "Riva TNT2 Ultra"); break;
                case PCI_DEVICE_ID_VANTA:
                    sprintf(device, "Riva TNT2 Vanta"); break;
                case PCI_DEVICE_ID_ULTRA:
                    sprintf(device, "Riva TNT2 Ultra"); break;
                case PCI_DEVICE_ID_INTEGRATED_TNT2:
                    sprintf(device, "Riva TNT2 Integrated"); break;
                case PCI_DEVICE_ID_GEFORCE:
                    sprintf(device, "GeForce256"); break;
                case PCI_DEVICE_ID_GEFORCEDDR:
                    sprintf(device, "GeForce256 DDR"); break;
                case PCI_DEVICE_ID_QUADRO:
                    sprintf(device, "Quadro"); break;
                case PCI_DEVICE_ID_GEFORCE2:
                    sprintf(device, "GeForce2 MX"); break;
                case PCI_DEVICE_ID_GEFORCE2_1:
                    sprintf(device, "GeForce2 MX variant 1"); break;
                case PCI_DEVICE_ID_GEFORCE2_3:
                    sprintf(device, "GeForce2 MX variant 3"); break;
                case PCI_DEVICE_ID_GEFORCE2GTS:
                    sprintf(device, "GeForce2 GTS"); break;
                case PCI_DEVICE_ID_GEFORCE2GTSDDR:
                    sprintf(device, "GeForce2 GTS DDR"); break;
                case PCI_DEVICE_ID_GEFORCE2GTS_2:
                    sprintf(device, "GeForce2 GTS variant 2"); break;
                case PCI_DEVICE_ID_QUADRO2:
                    sprintf(device, "Quadro2"); break;
                case PCI_DEVICE_ID_NVIDIA_CARD0:
                case PCI_DEVICE_ID_NVIDIA_CARD1:
                case PCI_DEVICE_ID_NVIDIA_CARD2:
                case PCI_DEVICE_ID_NVIDIA_CARD3:
                    sprintf(device, "NVIDIA card"); break;
                default:
                    sprintf(device, "Unknown");
            }
            len += sprintf(page+len, "------ Card Info ------\n");
            len += sprintf(page+len, "Model:        %s\n", device);
            len += sprintf(page+len, "IRQ:          %d\n", tnv->interrupt_line);
            break;
        }
    }

    if (!(pcibios_find_class(PCI_CLASS_BRIDGE_HOST << 8, 0, &bus, &func))) {
        
        pcibios_read_config_dword(bus, func, 0x04, &status);
        
        if (!(status & 0x00100000))
            return len;
        
        pcibios_read_config_byte(bus, func, 0x34, &cap_ptr);
        
        if (cap_ptr == 0x00)
            return len;

        do {
            pcibios_read_config_dword(bus, func, cap_ptr, &cap_id);

            if ((cap_id & 0xff) != 0x02)
                cap_ptr = (cap_id >> 8) & 0xff;

        } while (cap_ptr != 0x00 && (cap_id & 0xff) != 0x02);

        if (cap_ptr == 0x00)
        {
            len += sprintf(page+len, "------ PCI Card -------\n");
            return len;
        }

        len += sprintf(page+len, "------ AGP Info -------\n");
        
        pcibios_read_config_word(bus, func, PCI_VENDOR_ID, &vendor_id);
        pcibios_read_config_word(bus, func, PCI_DEVICE_ID, &device_id);
        
        switch (vendor_id) {
            case PCI_VENDOR_ID_AL:
                switch (device_id) {
                    case PCI_DEVICE_ID_AL_M1541_0:
                        sprintf(chipset, "Ali M1541"); break;
                    default:
                        sprintf(chipset, "Generic Ali");
                }
                break;
            case PCI_VENDOR_ID_AMD:
                switch (device_id) {
                    case PCI_DEVICE_ID_AMD_IRONGATE_0:
                        sprintf(chipset, "AMD Irongate"); break;
                    default:
                        sprintf(chipset, "Generic AMD");
                }
                break;
            case PCI_VENDOR_ID_INTEL:
                switch (device_id) {
                    case PCI_DEVICE_ID_INTEL_82443BX_0:
                        sprintf(chipset, "Intel 440BX"); break;
                    case PCI_DEVICE_ID_INTEL_82443LX_0:
                        sprintf(chipset, "Intel 440LX"); break;
                    case PCI_DEVICE_ID_INTEL_82443GX_0:
                        sprintf(chipset, "Intel 440GX"); break;
                    case PCI_DEVICE_ID_INTEL_810_0:
                        sprintf(chipset, "Intel i810"); break;
                    case PCI_DEVICE_ID_INTEL_810_E_0:
                        sprintf(chipset, "Intel i810e"); break;
                    case PCI_DEVICE_ID_INTEL_840_0:
                        sprintf(chipset, "Intel i840"); break;
                    default:
                        sprintf(chipset, "Generic Intel");
                }
                break;
            case PCI_VENDOR_ID_VIA:
                switch (device_id) {
                    case PCI_DEVICE_ID_VIA_8501_0:
                        sprintf(chipset, "Via MVP4"); break;
                    case PCI_DEVICE_ID_VIA_82C597_0:
                        sprintf(chipset, "Via VP3"); break;
                    case PCI_DEVICE_ID_VIA_82C598_0:
                        sprintf(chipset, "Via MVP3"); break;
                    case PCI_DEVICE_ID_VIA_82C691_0:
                        sprintf(chipset, "Via Apollo Pro"); break;
                    case PCI_DEVICE_ID_VIA_8371_0:
                        sprintf(chipset, "Via Apollo Pro KX133"); break;
                    case PCI_DEVICE_ID_VIA_8363_0:
                        sprintf(chipset, "Via Apollo Pro KT133"); break;
                    default:
                        sprintf(chipset, "Generic Via");
                }
                break;
            case PCI_VENDOR_ID_SI:
                    sprintf(chipset, "Generic SiS"); break;
            default:
                sprintf(chipset, "Unknown");
        }

        /* read in the card's capabilities and current setup */        
        pcibios_read_config_dword(bus, func, cap_ptr + 4, &status);
        pcibios_read_config_dword(bus, func, cap_ptr + 8, &command);

        /* determine if AGP is enabled, and if so, which driver
         * is being used.
         */
        if (command & 0x100)
        {
            len += sprintf(page+len, "AGP status:   Enabled\n");
            if (tnv->agpgart)
                len += sprintf(page+len, "AGP Driver:   AGPGART\n");
            else
                len += sprintf(page+len, "AGP Driver:   NVIDIA\n");
        } else {
            len += sprintf(page+len, "AGP status:   Disabled\n");
            len += sprintf(page+len, "AGP Driver:\n");
        }
        
        len += sprintf(page+len, "Bridge:       %s\n", chipset);

        len += sprintf(page+len, "SBA:          %s [%s]\n", 
                status & 0x00000200 ? "Supported" : "Unsupported",
                command & 0x00000200 ? "enabled" : "disabled");

        len += sprintf(page+len, "FW:           %s [%s]\n",
                status & 0x00000010 ? "Supported" : "Unsupported",
                command & 0x00000010 ? "enabled" : "disabled");

        len += sprintf(page+len, "Rates:        %s%s%s ",
                status & 0x00000004 ? "4x " : "",
                status & 0x00000002 ? "2x " : "",
                status & 0x00000001 ? "1x " : "");
       
        switch (command & 0x7)
        {
            case 4: 
                len += sprintf(page+len, "[%s]\n", "4x"); break;
            case 2: 
                len += sprintf(page+len, "[%s]\n", "2x"); break;
            case 1: 
                len += sprintf(page+len, "[%s]\n", "1x"); break;
            default:
                len += sprintf(page+len, "[-]\n"); break;
        }

        len += sprintf(page+len, "Registers:    0x%08x:0x%08x\n", status, command); 
    }

    return len;
}                               
#endif /* CONFIG_PROC_FS */



/***
 *** EXPORTS to rest of resman
 ***/

/*
 * Convert virtual (user or kernel) to "physical address"
 *   Here physical means the bus address, as we're just prepping
 *   to give it to dma
 */
int nv_vtop(
    nv_ioctl_memory_vtop_t *params,
    int                     copy_to_user
)
{
    void *v;
    int l;
    unsigned long p;		/* physical addr */
    u32 *physical_addresses;

    // page align everything

    v = NV_ALIGN_PTR_DOWN(params->buffer, NVHW_PAGE_SIZE);
    l = NV_ALIGN_UP(params->byte_length + (params->buffer - v), NVHW_PAGE_SIZE);

    for (physical_addresses = params->physical_addresses;
         l > 0;
         physical_addresses++)
    {
        // get the physical address and copy it out to caller's buffer
        p = (unsigned long) nv_get_phys_address((unsigned long) v);
            
        NV_DMSG((nv_state_t *) 0, "vtop: 0x%p to p 0x%08x", v, (unsigned int) p);
        if (copy_to_user)
            put_user(p, physical_addresses);
        else
            *physical_addresses = p;
        l -= PAGE_SIZE;
        v += PAGE_SIZE;
    }

    return 0;
}

void *nv_find_kernel_mapping(
    nv_state_t    *nv,
    void          *address
)
{
    nv_alloc_t *at;
    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) address, 
        (nv_alloc_t **) 0);
    if (at && at->page_table)
    {
        unsigned int page = ((unsigned long) address - at->vma->vm_start) >> 
                            PAGE_SHIFT;
        return __va(at->page_table[page]);
    }

    return (void *) 0;
}

/* virtual address to physical page address */
unsigned long 
nv_get_phys_address(unsigned long address) 
{
    unsigned long ret_addr;

    if (address > VMALLOC_START)
        address = VMALLOC_VMADDR(address);

    /* make sure we don't seg fault on a bad address */
    ret_addr = nvos_get_kvirt_address(address);
    if (ret_addr) ret_addr = __pa( (void *) ret_addr);

    return ret_addr;
}

/* allocate memory for DMA push buffers */
int
nv_alloc_pages(
    nv_state_t *nv,
    void **pAddress,
    unsigned int page_count,
    unsigned int agp_memory,
    unsigned int class,
    void **priv_data
)
{
    nv_alloc_t *at;
    int rm_status = 0;
    nv_linux_state_t *nvl = (nv_linux_state_t *) nv;

    NV_KMALLOC(at, sizeof(nv_alloc_t));
    if (at == NULL)
    {
        printk("NVRM: failed to allocate data structures for alloc pages\n");
        return rm_status;
    }

    osMemSet(at, 0, sizeof(nv_alloc_t));
    at->num_pages = NVHW_PAGES_TO_OS_PAGES(page_count);
    at->process_id = current->pid;
    at->next = nvl->alloc_queue;
    nvl->alloc_queue = at;

    if (agp_memory)
    {
        /* allocate agp-able memory */
        if (nv->agpgart)
        {
            NVRM_LOCK_SETUP();
            /* agpgart will allocate all of the underlying memory */
            NVRM_LOCK_RELEASE();
            rm_status = KernAllocAGPPages(pAddress, page_count, priv_data);
            NVRM_LOCK_GRAB();
            if (rm_status)
            {
                goto failed;
            }
            at->agp_mapping = *pAddress;
            at->priv_data = *priv_data;
        } else {
            /* use nvidia's nvagp support */
            at->page_table = nvos_malloc(page_count << PAGE_SHIFT);
            if (at->page_table == NULL)
            {
                goto failed;
            }
            *pAddress = (void *) at->page_table;
            rm_status = NvAllocAGPPages(nv->pdev, 
                                        pAddress,
                                        page_count, 
                                        class, 
                                        priv_data);
            if (rm_status)
            {
                goto failed;
            }
            at->class = class;
            at->priv_data = *priv_data;
        }
        nv->agp_buffers++;
    } else {
        /* allocate general system memory */
        at->page_table = nvos_malloc(page_count << PAGE_SHIFT);
        if (at->page_table == NULL)
        {
            goto failed;
        }
        *pAddress = (void *) at->page_table;
    }

    return rm_status;

failed:
    /* free any pages we may have allocated */
    if (at->page_table)
        nvos_free(at->page_table);

    /* remove our at from the queue */
    if (nvl->alloc_queue == at)
        nvl->alloc_queue = at->next;
    else
    {
        nv_alloc_t *prev = nvl->alloc_queue;
        while (prev && prev->next != at)
            prev = prev->next;
        if (prev && prev->next == at)
            prev->next = at->next;
    }

    /* free it */
    NV_KFREE(at);
    return -1;
}

int
nv_free_pages(
    nv_state_t *nv,
    void **pAddress,
    unsigned int page_count,
    unsigned int agp_memory,
    void **priv_data
)
{
    int rmStatus = 0;
    nv_alloc_t *at, *prev; 
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    if (agp_memory)
    {
        at = nvl_find_agp_alloc(nvl, (unsigned long) *pAddress, &prev);
        if (at == NULL)
        {
            printk("Couldn't find agp alloc for 0x%p\n", *pAddress);
            return -1;
        }

        if (nv->agpgart)
        {
            NVRM_LOCK_SETUP();
            NVRM_LOCK_RELEASE();
            rmStatus = KernFreeAGPPages(pAddress, priv_data);
            NVRM_LOCK_GRAB();
        } else {
            rmStatus = NvFreeAGPPages(nv->pdev, pAddress, priv_data);
            if (rmStatus == 0x0)
                nvos_free(at->page_table);
        }

        /* we may hold off on disabling agp until all buffers are freed */
        if (rmStatus == 0x0)
        {
            nv->agp_buffers--;
            if (!nv->agp_buffers && nv->agp_teardown)
                nv_agp_teardown(nv);
        }
    } else {
        at = nvl_find_alloc(nvl, (unsigned long) *pAddress, &prev);
        if (at == NULL)
        {
            printk("Couldn't find system alloc for 0x%p\n", *pAddress);
            return -1;
        }
        nvos_free(at->page_table);
    }

    /* remove and free our memory allocation structure */
    if (prev)
        prev->next = at->next;
    else
        nvl->alloc_queue = at->next;
    NV_KFREE(at);

    return rmStatus;
}

void nv_lock_init(void)
{
    spin_lock_init(&nv_lock);
    nv_locked = NULL;
    nv_locked_by = 0;
}

void nv_lock_grab(
    unsigned long *pflags
)
{
    spin_lock_irqsave(&nv_lock, *pflags);
    nv_locked = pflags;		// make sure osAllocMem() knows not to sleep
    if (in_interrupt())
        nv_locked_by = -1;
    else
        nv_locked_by = current->pid;
}

void nv_lock_release(
    unsigned long flags
)
{
    nv_locked = (unsigned long *) 0;
    nv_locked_by = 0;
    spin_unlock_irqrestore(&nv_lock, flags);
}

/*
** post the event
** Called from osNotifyEvent()
**
** We already hold the spinlock when we get in here...
*/
void nv_post_event(
    void     *file_void
)
{
    nv_linux_state_t *nvl;
    nv_file_private_t *nvfp;
    struct file *file = file_void;

    nvfp = file->private_data;
    nvl = NV_GET_NVL_FROM_FILEP(file);

    NV_DMSG(NV_STATE_PTR(nvl), "post");

    nvfp->any_fired_notifiers++;

    wake_up_interruptible(GET_EVENT_QUEUE(nvl));
}

/*
** post vblank specifically
** Called from osVblankNotify(), from VBlank()
*/
void nv_post_vblank(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = (nv_linux_state_t *) nv->os_state;
    NV_DMSG(nv, "post");

    if (nvl->waiting_for_vblank) 
        nvl->vblank_notifier++;

    wake_up_interruptible(GET_EVENT_QUEUE(nvl));
}


int
nv_agp_init(
    nv_state_t *nv,
    VOID **phys_start,
    VOID **linear_start,
    VOID *agp_limit,
    U032 config         /* passed in from XF86Config file */
)
{
    U032 status = 1;

    if (config & NVOS_AGP_CONFIG_AGPGART)
    {    
        NVRM_LOCK_SETUP();
        NVRM_LOCK_RELEASE();
        status = KernInitAGP(phys_start, linear_start, agp_limit);
        NVRM_LOCK_GRAB();
            
        /* if enabling agpgart was successfull, register it, 
         * and check about overrides 
         */
        if (status == 0)
        {
            nv->agpgart = 1;
            /* make sure we apply our overrides in this case */
            NvUpdateAGPConfig(nv->pdev);
        }
    }

    /* we're either explicitly not using agpgart, 
     * or trying to use agpgart failed 
     * make sure the user did not specify "use agpgart only"
     */
    if ( (nv->agpgart == 0) && (config & NVOS_AGP_CONFIG_NVAGP) )
    {
        /* make sure the user does not have agpgart loaded */
#if defined(KERNEL_2_4) && !defined(IA64)
		if ( inter_module_get("drm_agp") ) {
			inter_module_put("drm_agp");
#else
        void *sym_tmp;
        NV_GET_MODULE_SYMBOL(sym_tmp, 0, __MODULE_STRING(agp_enable));
        if (sym_tmp) {
#endif
            printk("NVRM: not using NVAGP, AGPGART is loaded!!\n");
        } else
            status = RmInitAGP(nv->pdev);
    }

    return status;
}

int
nv_agp_teardown(
    nv_state_t *nv
)
{
    U032 status = 1;

    /* if agp buffers are still in use, don't teardown just yet */
    if (nv->agp_buffers)
    {
        nv->agp_teardown = 1;
        return 0;
    }

    if (nv->agpgart)
    {
        NVRM_LOCK_SETUP();
        NVRM_LOCK_RELEASE();
        status = KernTeardownAGP();
        NVRM_LOCK_GRAB();
        nv->agpgart = 0;
    } else
        status = RmTeardownAGP(nv->pdev);

    nv->agp_teardown = 0;

    return status;
}

/* 
 * this takes a virtual address related to agp and returns the physical address
 * this part is a bit hacky.. nvagp expects a kernel-linear address mapping, 
 * but we don't have that. So, what's really passed to us is a "fake linear"
 * address, based at our page_table. nvl_find_agp_alloc() knows this, and will
 * return the 'at' that has the page_table. From there, we deconstruct this
 * 'fake address' to get a "page index" into the virtual address. We use this
 * index to grab a page from our page_table and return that physical address.
 * (if the linear mapping was real, this is the page that would have been
 * addressed by the virtual address. I know, confusing).
 */ 
int
nv_agp_translate_address(
    nv_state_t *nv,
    void       *base,
    U032        index,
    U032       *paddr
)
{
    nv_alloc_t *at;                                           

    /* find the 'at' */              
    at = nvl_find_agp_alloc(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) base, 0);
    if (at == NULL)
        return -1;

    /* get the physical address of this page */
    *paddr = (U032) ((NV_UINTPTR_T)at->page_table[index]);

    return 0x0;
}

/* For AGP chipsets such as the 460GX, the virtual address is not mapped
 * to the aperture address on the CPU.  This return returns the mapping so
 * that osGetAddressInfo will work.
 */
int
nv_agp_retrieve_mapping(
    nv_state_t    *nv,
    void          *address,
    unsigned long *agp_address
)
{
    nv_alloc_t *at;

    at = nvl_find_agp_alloc(NV_GET_NVL_FROM_NV_STATE(nv), (unsigned long) address, 0);
    if ((at == NULL) || (at->agp_mapping == NULL))
        return 1;

    *agp_address = (unsigned long)((NV_UINTPTR_T)at->agp_mapping);

    return 0x0;
}
